import Auth from "./src/helpers/auth";
import Utils from "./src/helpers/utils";
import { createWriteStream } from "fs";
import { utimes } from "fs/promises";
import { Readable } from "stream";
import { Semaphore } from "@inrixia/helpers";

const auth = await Auth()

const utils = new Utils(auth);

await utils.init()

Bun.serve({
    routes: {
        "/api/track/:id": async req => {
            const trackId = parseInt(req.params.id)

            const flac = Bun.file(`downloaded/${trackId}.flac`)

            if (await flac.exists()) {
                return new Response(flac, {
                    headers: {
                        "Content-Type": "audio/flac",
                        //"Content-Disposition": `attachment; filename="${trackId}.flac"`,
                        "Cache-Control": "public, max-age=31536000",
                        "ETag": trackId.toString(),
                    }
                })
            }

            const m4a = Bun.file(`downloaded/${trackId}.m4a`)

            if (await m4a.exists()) {
                return new Response(m4a, {
                    headers: {
                        "Content-Type": "audio/m4a",
                        //"Content-Disposition": `attachment; filename="${trackId}.m4a"`,
                        "Cache-Control": "public, max-age=31536000",
                        "ETag": trackId.toString(),
                    }
                })
            }

            const trackData = await utils.fetchTrack(trackId);

            const stream = await utils.downloadAudio(trackData);

            const format = trackData.manifestMimeType === "application/vnd.tidal.bts" ? "flac" : "m4a";

            return new Response(stream, {
                headers: {
                    "Content-Type": `audio/${format}`,
                    "Cache-Control": "public, max-age=31536000",
                    "ETag": trackId.toString(),
                }
            })
        },

        "/api/@me/tracks": async () => {
            const tracks = await utils.fetchTracks();

            return new Response(Bun.gzipSync(JSON.stringify(tracks)), {
                headers: {
                    "Content-Type": "application/json",
                    "Content-Encoding": "gzip",
                }
            })
        }
    },
    development: true,
    idleTimeout: 255
})

const tracks = await utils.fetchTracks() as {
    limit: number,
    offset: number,
    totalNumberOfItems: number,
    items: {
        created: string,
        item: {
            id: string
            title: string
            duration: number, // in seconds
            replayGain: number,
            peak: number,
            allowStreaming: boolean,
            streamReady: boolean,
            payToStream: boolean,
            adSupportedStreamReady: boolean,
            djReady: boolean,
            stemReady: boolean,
            streamStartDate: string,
            premiumStreamingOnly: boolean,
            trackNumber: number,
            volumeNumber: number, // discNumber
            version: string, // Slowed, Acoustic, Remix, etc
            popularity: number,
            copyright: string,
            bpm: null | number, // not available for all tracks
            url: string,
            isrc: string,
            editable: boolean,
            explicit: boolean,
            audioQuality: string,
            audioModes: "STEREO" | "DOLBY_ATMOS"[], // todo find other modes
            mediaMetadata: {
                tags: "DOLBY_ATMOS" | "LOSSLESS" | "HIRES_LOSSLESS"[], // todo find other tags
            },
            upload: boolean, // was uploaded by a user
            artist: {
                id: number,
                name: string,
                handle: null, // ?
                type: "MAIN", // todo find other types
                picture: string,
            },
            artists: [
                {
                    id: number,
                    name: string,
                    handle: null,
                    type: "MAIN" | "FEATURED",
                    picture: string,
                }
            ],
            album: {
                id: number,
                title: string,
                cover: string,
                vibrantColor: string,
                videoCover: null | string, // only available for some albums
            },
            mixes: {
                TRACK_MIX?: string,
            },
        }
    }[]
};


/* Remove all tracks that are unavailable*/
let unavailableTracks = tracks.items.filter(track => !track.item.streamReady);

/*for (const track of unavailableTracks) {
    await utils.fetch("/users/199235629/favorites/tracks/" + track.item.id, {
        method: "DELETE",
    })
    console.log("Removed from favorites", track.item.id)
}*/
/*
Bun.write("unavailable.json", JSON.stringify(unavailableTracks, null, 2))

function findDuplicateTracksByName(tracks: typeof tracks.items) {
    const seen = new Map<string, { count: number, originals: typeof tracks }>();

    for (const trackObj of tracks) {
        const title = trackObj.item.title;
        const normalized = title.trim().toLowerCase();

        if (!seen.has(normalized)) {
            seen.set(normalized, { count: 1, originals: [trackObj] });
        } else {
            const entry = seen.get(normalized)!;
            entry.count++;
            entry.originals.push(trackObj);
        }
    }

    // Filter to only include titles with duplicates
    return Array.from(seen.entries())
        .filter(([_, v]) => v.count > 1)
        .map(([normalizedTitle, data]) => ({
            normalizedTitle,
            count: data.count,
            duplicates: data.originals,
        }));
}

function findDuplicateTracksByISRC(tracks: typeof tracks.items) {
    const seen = new Map<string, { count: number, originals: typeof tracks }>();

    for (const trackObj of tracks) {
        const isrc = trackObj.item.isrc;

        // Skip if ISRC is missing or empty
        if (!isrc) continue;

        if (!seen.has(isrc)) {
            seen.set(isrc, { count: 1, originals: [trackObj] });
        } else {
            const entry = seen.get(isrc)!;
            entry.count++;
            entry.originals.push(trackObj);
        }
    }

    // Only return those with duplicates
    return Array.from(seen.entries())
        .filter(([_, v]) => v.count > 1)
        .map(([isrc, data]) => ({
            name: data.originals[0].item.title,
            isrc,
            count: data.count,
            duplicates: data.originals,
        }));
}


// Usage:
const duplicatesByISRC = findDuplicateTracksByISRC(tracks.items);

let duplicatesISRCText = {};

duplicatesByISRC.forEach(({ name, isrc, count, duplicates }) => {
    duplicatesISRCText[isrc] = { name, count, duplicates };
});

await Bun.write("duplicates.isrc.json", JSON.stringify(duplicatesISRCText, null, 2));


const duplicates = findDuplicateTracksByName(tracks.items);

let duplicatesNamesText = {};

duplicates.forEach(({ normalizedTitle, count, duplicates }) => {
    duplicatesNamesText[normalizedTitle] = { count, duplicates };
});

await Bun.write("duplicates.name.json", JSON.stringify(duplicatesNamesText, null, 2));
*/

const dlSemaphore = new Semaphore(1);

await Promise.all(tracks.items.map(async track => {
    dlSemaphore.with(async () => {
        try {
            const trackData = await utils.fetchTrack(parseInt(track.item.id));

            const stream = await utils.downloadAudio(trackData);

            const format = trackData.manifestMimeType === "application/vnd.tidal.bts" ? "flac" : "m4a";

            const writeStream = createWriteStream(`downloaded/${track.item.id}.${format}`, {
                flags: "w"
            });

            stream.pipe(writeStream);

            writeStream.on("finish", () => {
                console.log(`Downloaded ${track.item.id}.${format}`);
            });
        } catch (e) {
            console.error(`Failed to download ${track.item.id}`, e);
        }
    })
}))