import { credentialsProvider } from "@tidal-music/auth";
import { FlacStreamTagger, PictureType } from "flac-stream-tagger";
import { parse } from "dasha";

import { Readable } from "stream";

import { rejectNotOk } from "@inrixia/helpers";


import { makeDecipher } from "./decrypt";
import { fetchStream } from "./fetchStream";


export default class {
    private _authHeaders: { [key: string]: string }
    private _userId: string | undefined;
    private _userCountry: string;
    private _userLocale: string;

    constructor(authHeaders: { [key: string]: string }) {
        this._authHeaders = authHeaders;
        this._userCountry = "WW"; // Default to Worldwide
        this._userLocale = (process.env.LC_ALL ||
            process.env.LANG ||
            process.env.LANGUAGE ||
            Intl.DateTimeFormat().resolvedOptions().locale ||
            "en_US").replaceAll("-", "_").split(".")[0];
    }

    async fetch(url: string, opts: { [key: string]: any } = {}) {
        const parsedUrl = new URL(`https://desktop.tidal.com/v1${url}`);

        parsedUrl.searchParams.set("countryCode", this._userCountry);
        parsedUrl.searchParams.set("locale", this._userLocale);
        parsedUrl.searchParams.set("deviceType", "DESKTOP");

        const response = rejectNotOk(await fetch(parsedUrl, {
            headers: {
                ...this._authHeaders,
                "x-tidal-token": "mhPVJJEBNRzVjr2p",
            },
            referrer: "https://desktop.tidal.com/",
            ...opts
        }))

        return await response.json() as Object;
    }

    async init() {
        const creds = await credentialsProvider.getCredentials()

        this._userId = creds.userId

        const { countryCode } = await this.fetch("/country") as { countryCode: string }
        this._userCountry = countryCode;
    }

    async fetchTrack(id: number) {
        const audio = await this.fetch(`/tracks/${id}/playbackinfo?audioquality=HI_RES_LOSSLESS&playbackmode=STREAM&assetpresentation=FULL`)

        return audio as {
            trackId: number,
            audioPresentation: string,
            audioMode: string,
            audioQuality: string,
            manifestMimeType: string,
            manifestHash: string,
            manifest: string,
            albumReplayGain: number,
            albumPeakAmplitude: number,
            trackReplayGain: number,
            trackPeakAmplitude: number,
            bitDepth: number,
            sampleRate: number,
            status?: number,
        }
    }

    async fetchTracks() {
        const tracks = await this.fetch(`/users/${this._userId}/favorites/tracks?offset=0&limit=10000&order=DATE&orderDirection=DESC`);

        return tracks;
    }

    async downloadAudio(trackData: {
        manifestMimeType: string,
        manifest: string,
        trackId: number,
    }) {

        switch (trackData.manifestMimeType) {
            case "application/vnd.tidal.bts": {
                const parsedManifest = JSON.parse(Buffer.from(trackData.manifest, "base64").toString("utf-8")) as {
                    "mimeType": string,
                    "codecs": string,
                    "encryptionType": string,
                    "keyId": string,
                    "urls": string[]
                }

                const stream = Readable.from(fetchStream(parsedManifest.urls, {
                    decipher: makeDecipher(parsedManifest)
                }))

                if (parsedManifest.mimeType !== "audio/flac") return stream;

                return stream.pipe(
                    new FlacStreamTagger(await this.makeTags(trackData.trackId))
                );
            }
            case "application/dash+xml": {
                const parsedManifest = await parse(Buffer.from(trackData.manifest, "base64").toString("utf-8"), "https://sp-ad-cf.audio.tidal.com")

                return Readable.from(fetchStream(parsedManifest.tracks.all[0].segments.map((s) => s.url)))
            }

            default: {
                throw new Error(`Unsupported Stream Info manifest mime type: ${trackData.manifestMimeType}`);
            }

        }
    }

    async makeTags(id: number) {
        const trackData = await this.fetch(`/tracks/${id}`) as {
            title: string,
            replayGain: number,
            peak: number,
            streamStartDate: string,
            trackNumber: number,
            volumeNumber: number, // discNumber
            copyright: string,
            bpm: number,
            url: string, // comment
            isrc: string,
            artists: [
                {
                    name: string,
                }
            ],
            album: {
                id: number,
                title: string,
                cover: string,
            },
            status?: string,
        };

        const albumData = await this.fetch(`/albums/${trackData.album.id}`) as {
            title: string,
            numberOfTracks: number,
            releaseDate: string,
            cover: string,
            upc: string,
            artists: [
                {
                    name: string,
                }
            ],
        };

        const albumArtReq = await fetch(`https://resources.tidal.com/images/${albumData.cover.replaceAll("-", "/") || ""}/1280x1280.jpg`)

        const picture = {
            buffer: Buffer.from(await albumArtReq.arrayBuffer()),
            type: PictureType.FrontCover,
        }

        return {
            tagMap: {
                title: trackData?.title || "",
                trackNumber: (trackData?.trackNumber || 0).toString(),
                discNumber: (trackData?.volumeNumber || 0).toString(),
                bpm: (trackData?.bpm || 0).toString(),
                date: trackData?.streamStartDate || "",
                copyright: trackData?.copyright || "",
                REPLAYGAIN_TRACK_GAIN: (trackData?.replayGain || 0).toString(),
                REPLAYGAIN_TRACK_PEAK: (trackData?.peak || 0).toString(),
                comment: trackData?.url || "",
                isrc: trackData?.isrc || "",
                upc: albumData?.upc || "",
                artist: trackData?.artists.map((a) => a.name) || [],
                album: albumData?.title || "",
                albumArtist: albumData?.artists.map((a) => a.name) || [],
                totalTracks: (albumData?.numberOfTracks || 0).toString(),
                year: albumData?.releaseDate.split("-")[0] || "",
            },
            picture
        }
    }
}